#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "ATmega88PA.h"
#include "device.h"
#include "display.h"

#define CLK_PULSE CLK_HI; CLK_LOW
#define LE_PULSE LE_HI; LE_LOW
#define DRIVER_1 4
#define DRIVER_2 0
#define DRIVER_3 3
#define DRIVER_4 1
#define DRIVER_5 2

static uint8_t driver = 0;
static uint8_t column = 0;
static uint8_t sw_led = 0;
static uint8_t dot = 0;

static const uint8_t font [][5] PROGMEM = {
    {0x3E,0x41,0x41,0x41,0x3E}, // 0
    {0x00,0x42,0x7F,0x40,0x00}, // 1
    {0x62,0x51,0x49,0x49,0x46}, // 2
    {0x22,0x41,0x49,0x49,0x36}, // 3
    {0x18,0x14,0x12,0x7F,0x10}, // 4
    {0x27,0x45,0x45,0x45,0x39}, // 5
    {0x3C,0x4A,0x49,0x49,0x30}, // 6
    {0x01,0x71,0x09,0x05,0x03}, // 7
    {0x36,0x49,0x49,0x49,0x36}, // 8
    {0x06,0x49,0x49,0x29,0x1E}, // 9
    {0x7E,0x11,0x11,0x11,0x7E}, // A
    {0x7F,0x49,0x49,0x49,0x36}, // B
    {0x3E,0x41,0x41,0x41,0x22}, // C
    {0x7F,0x41,0x41,0x41,0x3E}, // D
    {0x7F,0x49,0x49,0x49,0x41}, // E
    {0x7F,0x09,0x09,0x09,0x01}, // F
    {0x3E,0x41,0x49,0x49,0x3A}, // G
    {0x7F,0x08,0x08,0x08,0x7F}, // H
    {0x00,0x41,0x7F,0x41,0x00}, // I
    {0x20,0x40,0x40,0x40,0x3F}, // J
    {0x7F,0x08,0x14,0x22,0x41}, // K
    {0x7F,0x40,0x40,0x40,0x40}, // L
    {0x7F,0x02,0x04,0x02,0x7F}, // M
    {0x7F,0x04,0x08,0x10,0x7F}, // N
    {0x3E,0x41,0x41,0x41,0x3E}, // O
    {0x7F,0x09,0x09,0x09,0x06}, // P
    {0x7F,0x09,0x19,0x29,0x46}, // R
    {0x26,0x49,0x49,0x49,0x32}, // S
    {0x01,0x01,0x7F,0x01,0x01}, // T
    {0x3F,0x40,0x40,0x40,0x3F}, // U
    {0x7F,0x20,0x10,0x20,0x7F}, // W
    {0x07,0x18,0x60,0x18,0x07}, // V
    {0x63,0x14,0x08,0x14,0x63}, // X
    {0x07,0x08,0x70,0x08,0x07}, // Y
    {0x61,0x51,0x49,0x45,0x43}, // Z
    {0x00,0x00,0x36,0x00,0x00}, // :
    {0x00,0x00,0x60,0x00,0x00}, // .
    {0x00,0x06,0x09,0x09,0x06}, // degree
    {0x08,0x08,0x00,0x00,0x00}, // minus
    {0x02,0x01,0x51,0x09,0x06}  // ?
};

#define BRIGHTNESS(n) pgm_read_byte(&(lightness_correction[(n)]))

static int8_t brightness = 0;

static const uint8_t lightness_correction[32] PROGMEM = {
    0xFF,0xFE,0xFD,0xFC,0xFB,0xFA,0xF8,0xF6,
    0xF3,0xF0,0xED,0xE9,0xE4,0xDF,0xD9,0xD3,
    0xCC,0xC5,0xBC,0xB3,0xA9,0x9F,0x93,0x87,
    0x79,0x6B,0x5C,0x4B,0x3A,0x27,0x14,0x00
};

static uint8_t buffer[25] = {
    0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00
};

void DISPLAY_Clear(void) {
    uint8_t i = 24;
    do {
        buffer[i] = 0x00;
    } while(i--);
}

void DISPLAY_Column(uint8_t data, uint8_t offset) {
    buffer[offset] = data;
}

void DISPLAY_Print(uint8_t symbol, uint8_t offset) {
    uint8_t i = 4;
    do {
        buffer[offset+i] |= pgm_read_byte(&(font[symbol][i]));
    } while(i--);
}

void DISPLAY_SetBrightness(uint8_t level) {
    TIMER0_PWMB_PULSE_WIDTH(BRIGHTNESS(level + 11));
    brightness = level;
}

void DISPLAY_CheckBrightness(void) {
    ADC_START_CONVERSION;
}

void DISPLAY_ButtonBacklight(uint8_t backlight) {
    sw_led = backlight;
}

void DISPLAY_Dot_ON(void) { dot = 1; }
void DISPLAY_Dot_OFF(void) { dot = 0; }

static void send_column(uint8_t driver, uint8_t column) {
    uint8_t disp1 = buffer[column];
    uint8_t disp2 = buffer[column+5];
    uint8_t disp3 = buffer[column+10];
    uint8_t disp4 = buffer[column+15];
    uint8_t disp5 = buffer[column+20];
    switch(driver) {
        case DRIVER_1:
            (disp1 & 0x10) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp1 & 0x04) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp1 & 0x20) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp1 & 0x40) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (sw_led & SW1) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp1 & 0x01) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp1 & 0x02) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x04) ? SDI_HI : SDI_LOW; CLK_PULSE;
            break;
        case DRIVER_2:
            (disp2 & 0x08) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x08) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp3 & 0x10) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp3 & 0x20) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp3 & 0x40) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x40) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x20) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x10) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp1 & 0x08) ? SDI_HI : SDI_LOW; CLK_PULSE;
            break;
        case DRIVER_3:
            (disp3 & 0x04) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x02) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp2 & 0x01) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (sw_led & SW2) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (sw_led & SW3) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp3 & 0x01) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp3 & 0x02) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp4 & 0x04) ? SDI_HI : SDI_LOW; CLK_PULSE;
            break;
        case DRIVER_4:
            (disp4 & 0x08) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp5 & 0x10) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp5 & 0x20) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp5 & 0x40) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp4 & 0x40) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp4 & 0x20) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp4 & 0x10) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp3 & 0x08) ? SDI_HI : SDI_LOW; CLK_PULSE;
            break;
        case DRIVER_5:
            (disp5 & 0x04) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp4 & 0x02) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp4 & 0x01) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (sw_led & SW4) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp5 & 0x01) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp5 & 0x02) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (dot) ? SDI_HI : SDI_LOW; CLK_PULSE;
            (disp5 & 0x08) ? SDI_HI : SDI_LOW; CLK_PULSE;
            break;
    }
}

void DISPLAY_Init(void) {
    //make sure display is off
    A1_A2_A3_A4_A5_OFF;
    OE_HI;
    send_column(0, 0);
    send_column(1, 0);
    send_column(2, 0);
    send_column(3, 0);
    send_column(4, 0);
    LE_PULSE;
    //Timer0 configuration (PWMB = OE)
    TIMER0_FAST_PWM_FULL_RANGE_MODE;
    TIMER0_PWMB_CLEAR_ON_COMPARE_MATCH;
    TIMER0_CLEAR_OVERFLOW_INTERRUPT_FLAG;
    TIMER0_OVERFLOW_INTERRUPT_ENABLE;
    TIMER0_PWMB_PULSE_WIDTH(0xFF);
    TIMER0_CLOCK_PRESCALER_8_START;
    //ADC configuration (light sensor on channel 2)
    ADC_VOLTAGE_REFERENCE_AVCC;
    ADC_LEFT_ADJUST_RESULT;
    ADC_SELECT_CHANNEL_2;
    ADC_DIGITAL_INPUT_ADC2_DISABLE;
    ADC_PRESCALER_128;
    ADC_ENABLE;
    ADC_CLEAR_INTERRUPT_FLAG;
    ADC_INTERRUPT_ENABLE;
}

ISR(ADC_vect) {
    uint8_t adc = (ADCH >> 2);
    uint8_t level = (brightness << 2);
    uint8_t hysteresis = 0x02;
    if(adc < (level - hysteresis)) { --brightness; }
    if(adc > (level + hysteresis)) { ++brightness; }
    DISPLAY_SetBrightness(brightness);
}

ISR(TIMER0_OVF_vect) {
    send_column(driver, column);
    if(++driver > 4) {
        driver = 0;
        A1_A2_A3_A4_A5_OFF;
        LE_PULSE;
        switch(column) {
            default:
            case 0: A1_ON; break;
            case 1: A2_ON; break;
            case 2: A3_ON; break;
            case 3: A4_ON; break;
            case 4: A5_ON; break;
        }
        if(++column > 4) {
            column = 0;
        }
    }
}
